/*
 * Copyright 2019 Bloomberg Finance LP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <buildboxcasd_hardlinkstager.h>

#include <buildboxcommon_exception.h>
#include <buildboxcommon_fileutils.h>
#include <buildboxcommon_logging.h>
#include <buildboxcommon_protos.h>
#include <buildboxcommon_timeutils.h>

#include <chrono>
#include <dirent.h>
#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

using namespace buildboxcasd;
using namespace build::bazel::remote::execution::v2;

HardLinkStager::HardLinkStager(LocalCas *cas_storage) : FileStager(cas_storage)
{
}

HardLinkStager::HardLinkStagedDirectory::HardLinkStagedDirectory(
    const std::string &path)
    : d_staged_path(path)
{
}

HardLinkStager::HardLinkStagedDirectory::~HardLinkStagedDirectory()
{
    unstage(d_staged_path);
}

std::unique_ptr<FileStager::StagedDirectory>
HardLinkStager::stage(const Digest &root_digest, const std::string &path)
{
    const bool path_is_existing_directory =
        buildboxcommon::FileUtils::is_directory(path.c_str());

    if (path_is_existing_directory &&
        !buildboxcommon::FileUtils::directory_is_empty(path.c_str())) {
        BUILDBOXCOMMON_THROW_EXCEPTION(std::invalid_argument,
                                       "Could not stage "
                                           << toString(root_digest) << " as ["
                                           << path << "] is not empty");
    }
    else {
        buildboxcommon::FileUtils::create_directory(path.c_str());
    }

    try {
        stageDirectory(root_digest, path);

        // userchroot doesn't allow the chroot directory to be more permissive
        // than 0755.
        // FIXME: To allow non-userchroot clients to have a staging directory
        // with mode 0777, this restriction should be communicated by the
        // client instead of being hardcoded here.
        chmod(path.c_str(), 0755);
    }
    catch (const std::exception &e) {
        BUILDBOX_LOG_DEBUG("Error staging root digest " << root_digest
                                                        << " in [" << path
                                                        << "]: " << e.what());

        // If the staging fails, we roll back the stage directory to the
        // its status before the call:
        unstage(path);

        throw;
    }

    // This object will keep the directory staged until it's destruction:
    return std::make_unique<HardLinkStagedDirectory>(path);
}

void HardLinkStager::unstage(const std::string &path)
{
    try {
        buildboxcommon::FileUtils::clear_directory(path.c_str());
    }
    catch (const std::system_error &e) {
        BUILDBOX_LOG_ERROR("Error unstaging " << path << ": " << e.what());
    }
}

void HardLinkStager::stageDirectory(const Digest &digest,
                                    const std::string &path)
{

    const Directory directory = getDirectory(digest);

    createDirectory(path);

    for (const FileNode &file : directory.files()) {
        const std::string stage_path = path + "/" + file.name();
        stageFile(file, stage_path);
    }

    for (const DirectoryNode &subdirectory : directory.directories()) {
        const std::string subdirectory_path = path + "/" + subdirectory.name();
        stageDirectory(subdirectory.digest(), subdirectory_path);
    }

    for (const SymlinkNode &symlink : directory.symlinks()) {
        const std::string symlink_path = path + "/" + symlink.name();
        stageSymLink(symlink.target(), symlink_path);
    }
}

void HardLinkStager::applyFileProperties(const FileNode &file,
                                         const std::string &path)
{
    for (const NodeProperty &property : file.node_properties()) {
        if (property.name() == "MTime") {
            const std::string timestamp = property.value();
            if (!timestamp.empty()) {
                const std::chrono::system_clock::time_point timepoint =
                    buildboxcommon::TimeUtils::parse_timestamp(timestamp);
                buildboxcommon::FileUtils::set_file_mtime(path.c_str(),
                                                          timepoint);
            }
            else {
                BUILDBOXCOMMON_THROW_EXCEPTION(
                    std::runtime_error,
                    "Timestamp for " << path << " is an empty string");
            }
        }
        else {
            BUILDBOXCOMMON_THROW_EXCEPTION(
                std::runtime_error, "Unsupported FileNode property "
                                        << property.name() << " with value "
                                        << property.value());
        }
    }
}

void HardLinkStager::stageFile(const FileNode &file, const std::string &path)
{
    const std::string cas_path = d_cas_storage->path(file.digest());

    // if we require properties, we must copy instead of hardlinking
    bool can_hardlink = (file.node_properties_size() == 0);

    if (can_hardlink) {
        hardLink(cas_path.c_str(), path.c_str());
        // (!) TODO: ensure links are read only
    }
    else {
        buildboxcommon::FileUtils::copy_file(cas_path.c_str(), path.c_str());
        // try to apply stored properties
        applyFileProperties(file, path);
    }

    if (file.is_executable()) {
        try {
            buildboxcommon::FileUtils::make_executable(path.c_str());
        }
        catch (const std::runtime_error &e) {
            BUILDBOX_LOG_ERROR("Could not give executable permission to "
                               << path << ": " << e.what());
            throw e;
        }
    }
}

void HardLinkStager::stageSymLink(const std::string &target,
                                  const std::string &link_path)
{
    if (target.find("//") != std::string::npos ||
        target.find("/./") != std::string::npos) {
        BUILDBOXCOMMON_THROW_EXCEPTION(
            std::invalid_argument,
            "Found forbidden `/./` or `//` in symlink target path");
    }

    const int link_status = symlink(target.c_str(), link_path.c_str());

    if (link_status != 0) {
        BUILDBOXCOMMON_THROW_SYSTEM_EXCEPTION(
            std::system_error, errno, std::system_category,
            "Could not create symlink " << link_path << " -> " << target);
    }
}

void HardLinkStager::hardLink(const std::string &target,
                              const std::string &link_path)
{
    const int link_status = link(target.c_str(), link_path.c_str());
    // (old path, new path)

    if (link_status != 0) {
        BUILDBOXCOMMON_THROW_SYSTEM_EXCEPTION(
            std::system_error, errno, std::system_category,
            "Could not create hard link " << link_path << " -> " << target);
    }
}

void HardLinkStager::createDirectory(const std::string &path)
{
    try {
        buildboxcommon::FileUtils::create_directory(path.c_str());
        // The umask affects the permission of the created directory.
        // Explicitly set it to 0777 to allow write access by clients and
        // reduce non-determinism in executed actions.
        chmod(path.c_str(), 0777);
    }
    catch (const std::system_error &e) {
        BUILDBOX_LOG_ERROR("Could not create directory " << path << ": "
                                                         << e.what());
        throw e;
    }
}

Directory HardLinkStager::getDirectory(const Digest &digest)
{
    const std::string blob_data = d_cas_storage->readBlob(digest);

    Directory directory;
    if (directory.ParseFromString(blob_data)) {
        return directory;
    }

    BUILDBOXCOMMON_THROW_EXCEPTION(
        std::runtime_error,
        "Could not parse Directory from blob with digest: " << digest);
}
