﻿/*
 * Copyright 2019 Bloomberg Finance LP
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef INCLUDED_BUILDBOXCASD_LRULOCALCAS_H
#define INCLUDED_BUILDBOXCASD_LRULOCALCAS_H

#include <atomic>
#include <chrono>
#include <condition_variable>
#include <mutex>
#include <thread>

#include <buildboxcasd_localcas.h>
#include <buildboxcommon_client.h>
#include <buildboxcommon_protos.h>

namespace buildboxcasd {

class LruLocalCas final : public LocalCas {
    /* A LocalCas wrapper implementing LRU expiry of blobs. */
  public:
    LruLocalCas(std::unique_ptr<LocalCas> base_cas, int64_t quota_low,
                int64_t quota_high, int64_t reserved_space);

    ~LruLocalCas() override;

    void protectActiveBlobs(blob_time_type start_time);

    bool hasBlob(const Digest &digest) override;

    bool writeBlob(const Digest &digest, const std::string &data) override;

    bool moveBlob(const Digest &digest, const std::string &path) override;

    std::string readBlob(const Digest &digest, size_t offset,
                         size_t length) override;

    std::string readBlob(const Digest &digest) override;

    bool deleteBlob(const Digest &digest) override;

    void listBlobs(digest_time_callback_type callback) override;

    std::string path(const Digest &digest) override;

    const std::string &path() override;

    buildboxcommon::TemporaryDirectory createStagingDirectory() override;

    buildboxcommon::TemporaryFile createTemporaryFile() override;

    int64_t getDiskUsage() override;

    int64_t getDiskQuota() override;

    int64_t getAvailableDiskSpace() override;

    void verifyDiskUsage();

    LocalCas *getBase();

  private:
    std::unique_ptr<LocalCas> d_base_cas;
    std::atomic<int64_t> d_disk_usage;
    std::atomic<int64_t> d_effective_quota_low;
    std::atomic<int64_t> d_effective_quota_high;
    const int64_t d_configured_quota_low;
    const int64_t d_configured_quota_high;
    const int64_t d_reserved_space;
    bool d_protect_active_blobs;
    blob_time_type d_protect_active_time;
    std::mutex d_cleanup_mutex;

    std::thread d_disk_monitor_thread;

    // Condition variable to stop disk monitor thread
    std::condition_variable d_stop_cv;
    std::mutex d_stop_mutex;
    bool d_stop;

    int64_t calculateDiskUsage();

    void updateDiskQuota();

    void diskMonitorThread();

    void cleanup();

    /* This method takes the size of a blob and a `write_function` that
     * has the purpose of adding it to the CAS and must return `true`
     * if the contents of the CAS changed. (I.e. returning `false` when trying
     * to add a blob that is already stored.)
     *
     * If there is space available, `write_function` will be invoked and,
     * according to the truth value returned from it, the disk usage will be
     * updated accordingly.
     *
     */
    bool tryToAddBlobAndUpdateDiskUsage(
        google::protobuf::int64 blob_size,
        const std::function<bool()> &add_blob_function);
};

} // namespace buildboxcasd

#endif // INCLUDED_BUILDBOXCASD_LRULOCALCAS_H
